LIBRARY IEEE;
USE IEEE.STD_LOGIC_1164.ALL;
USE IEEE.NUMERIC_STD.ALL;
USE IEEE.STD_LOGIC_UNSIGNED.ALL;
USE IEEE.STD_LOGIC_ARITH.ALL;

USE WORK.blake_pak.ALL;

ENTITY ip_frame_signature IS
    PORT (
        clk: IN STD_LOGIC;
        --ulazni interfejs
        start_in: IN STD_LOGIC;
        data_vld_in: IN STD_LOGIC;
        data_in: IN STD_LOGIC_VECTOR(7 DOWNTO 0);
        ready_to_accept: OUT STD_LOGIC;
        --izlazni interfejs
        start_out: OUT STD_LOGIC;
        output_enable: OUT STD_LOGIC;
        data_out: OUT STD_LOGIC_VECTOR(7 DOWNTO 0)
    );
END ip_frame_signature;

ARCHITECTURE RTL OF ip_frame_signature IS
    --vrednost most significant 4 bita prvog bajta IP hedera za IPv4 protokol
    CONSTANT VERSION_IPV4: STD_LOGIC_VECTOR(3 DOWNTO 0)  := x"4";-- 4 means it is IPv4 type of protocol
    --duzina IP hedera je 20 bajtova
    CONSTANT HEDER_LENGTH: INTEGER := 20;

    COMPONENT blake_pipe IS
       PORT (
           clk: IN STD_LOGIC; 

           h: IN  STD_LOGIC_VECTOR (WIDTHx8-1 DOWNTO 0);
           m: IN  STD_LOGIC_VECTOR (WIDTHx16-1 DOWNTO 0);
           s: IN  STD_LOGIC_VECTOR (WIDTHx4-1 DOWNTO 0);
           t: IN  STD_LOGIC_VECTOR (WIDTHx2-1 DOWNTO 0);
           inEN: IN  STD_LOGIC;
     
           hash: OUT STD_LOGIC_VECTOR (WIDTHx8-1 DOWNTO 0);
           outEN: OUT  STD_LOGIC
    );
    END COMPONENT;

    COMPONENT padding_generation
    GENERIC(
        HEDER_LENGTH : INTEGER := 20
    );
    PORT (
        clk: IN STD_LOGIC;
        --ulazni interfejs
        total_length: IN STD_LOGIC_VECTOR(15 DOWNTO 0);
        start_padding: IN STD_LOGIC;
        --izlazni interfejs
        packet_length: OUT STD_LOGIC_VECTOR(15 DOWNTO 0);
        pading_length: OUT STD_LOGIC_VECTOR(6 DOWNTO 0);
        two_blake: OUT STD_LOGIC;
        blake_en: OUT STD_LOGIC;
        blake_cnt_max: OUT STD_LOGIC_VECTOR(6 DOWNTO 0);
        padding: OUT STD_LOGIC_VECTOR(7 DOWNTO 0);
        padding_valid: OUT STD_LOGIC
    );
    END COMPONENT;
    
    COMPONENT ip_frame_storing
    PORT (
        clk: IN STD_LOGIC;
        --ulazni interfejs
        start_in: IN STD_LOGIC;
        data_vld_in: IN STD_LOGIC;
        data_in: IN STD_LOGIC_VECTOR(7 DOWNTO 0);
        total_new_length: IN STD_LOGIC_VECTOR(15 DOWNTO 0);
        cnt: IN STD_LOGIC_VECTOR(10 DOWNTO 0);
        pading_length: IN STD_LOGIC_VECTOR(6 DOWNTO 0);
        --izlazni interfejs
        start_out: OUT STD_LOGIC;
        output_enable: OUT STD_LOGIC;
        data_out: OUT STD_LOGIC_VECTOR(7 DOWNTO 0)
    );
    END COMPONENT;
    
    COMPONENT addressable_shift_register IS
    GENERIC(
        DATA_SIZE : INTEGER := 20;
        MAX_DELAY : INTEGER := 20;  
        ADDR_LENGTH : INTEGER := 5      
    );
    PORT (
        clk: IN STD_LOGIC;
        data_in: IN STD_LOGIC_VECTOR(DATA_SIZE-1 DOWNTO 0);
        address: IN STD_LOGIC_VECTOR(ADDR_LENGTH-1 DOWNTO 0);
        data_out: OUT STD_LOGIC_VECTOR(DATA_SIZE-1 DOWNTO 0)

    );
    END COMPONENT;

    --cnt je brojac koji broji pristigle podatke. 
    --cnt_d je vrednost cnt brojaca zakasnjena za jedan takt.
    --cnt_2d je vrednost cnt brojaca zakasnjena za dva takta.
    --cnt_3d je vrednost cnt brojaca zakasnjena za tri takta.
    SIGNAL cnt,cnt_d, cnt_2d, cnt_3d  : STD_LOGIC_VECTOR(10 DOWNTO 0) := (OTHERS => '0');
    --vrednost dokle broji brojac. Ova vrednost zavisi od duzine poruke. Ako je duzina poruke na primer
    --1000 brojac ce brojati do 999. Informaciju do koje vrednosti se broji se dobija iz hedera tacnije
    --iz bajtova koji definisu ukupnu duzinu poruke pri cemu se ta vrednost umanji za jedan.
    SIGNAL cnt_upper_limit            : STD_LOGIC_VECTOR(10 DOWNTO 0) := (OTHERS => '0');
    --Vrednost countera koja je potrebna kako bi se insertovao busy signal odnosno vrednost signala ready_to_accept
    --oborila na nulu. Vrednost linita iznosi ukupna duzina poruke umanjena za dva bajta.
    SIGNAL cnt_upper_limit_busy       : STD_LOGIC_VECTOR(10 DOWNTO 0) := (OTHERS => '0');

    
    --Vrednost signala state_of_receiving_byte moze biti nula(state_MSB) ili jedan(state_LSB). Masina za racunanje check sume
    --shodno (https://en.wikipedia.org/wiki/IPv4_header_checksum) racunanje vrsi sa 16bitnim podacima. Nas ulaz je osmobitni tako
    --da je potrebno spakovati dva ulazna podatka u jedan 16-bitni. Kada je state_of_receiving_byte jednak state_MSB ulaz se pakuje
    --na gornjih 8 bita 16bitne poruke a kada je state_of_receiving_byte jednak state_LSB ulaz se pakuje na donjih 8 bita 16bitne poruke.
    CONSTANT state_MSB                : STD_LOGIC := '0';
    CONSTANT state_LSB                : STD_LOGIC := '1';
    SIGNAL state_of_receiving_byte    : STD_LOGIC := state_MSB;
    
    --pomocni signal masine za checkum koji cuva gornjih osam bita
    SIGNAL latched_msb_data           : STD_LOGIC_VECTOR(7 DOWNTO 0) := (OTHERS => '0');

    --pomocni signali za racunanje checksume(process 263 linija)
    SIGNAL check_sum_aux_1            : STD_LOGIC_VECTOR(16 DOWNTO 0) := (OTHERS => '0');
    SIGNAL check_sum_aux_2            : STD_LOGIC_VECTOR(15 DOWNTO 0) := (OTHERS => '0');
    SIGNAL check_sum                  : STD_LOGIC_VECTOR(15 DOWNTO 0) := (OTHERS => '0');

    --Zakasnjeni ulazni podaci za jedan, dva i tri takta respektivno.
    SIGNAL data_in_d                  : STD_LOGIC_VECTOR(7 DOWNTO 0) := (OTHERS => '0');
    SIGNAL data_in_2d                 : STD_LOGIC_VECTOR(7 DOWNTO 0) := (OTHERS => '0');
    SIGNAL data_in_3d                 : STD_LOGIC_VECTOR(7 DOWNTO 0) := (OTHERS => '0');

    --Ukoliko ulazni paket nije IPv4 baziran ili checksumu nije validna vrenost invalid_frame ce postati logicka jedinica.
    --Ovaj signal zahteva promenljivo kasnjenje u zavisnosti od paddinga pa je u tu svrhu koriscen adresibilan sift registar.
    SIGNAL invalid_frame              : STD_LOGIC := '0';
    SIGNAL invalid_frame_int_aux      : STD_LOGIC_VECTOR(0 DOWNTO 0) := (OTHERS => '0');
    SIGNAL invalid_frame_aux          : STD_LOGIC_VECTOR(0 DOWNTO 0) := (OTHERS => '0');
    SIGNAL address_shift_reg          : STD_LOGIC_VECTOR(6 DOWNTO 0) := (OTHERS => '0');
    SIGNAL address_shift_reg_aux      : STD_LOGIC_VECTOR(6 DOWNTO 0) := (OTHERS => '0');
    
    --ukupna duzina ulazne poruke ukljucujuci heder
    SIGNAL total_length               : STD_LOGIC_VECTOR(15 DOWNTO 0) := (OTHERS => '0');
    --ukupna duzina novoformirane poruke ukljucujuci heder
    SIGNAL total_new_length           : STD_LOGIC_VECTOR(15 DOWNTO 0) := (OTHERS => '0');
    
    --Ako je frejm validan donja dva signala ucestvuju u pripremi podataka za blejk modul. packet_data i packet valid definisu
    --payload podatke ip frejma koji ucestvuju pri racunanju hash-a buduci da ip heder ne ucestvuje u obradi podataka.
    SIGNAL packet_valid               : STD_LOGIC := '0'; 
    SIGNAL packet_data                : STD_LOGIC_VECTOR(7 DOWNTO 0) := (OTHERS => '0');  

    --zakasnjen ulazni valid za jedan, dva i tri takta respektivno.    
    SIGNAL data_vld_in_d              : STD_LOGIC := '0';
    SIGNAL data_vld_in_2d             : STD_LOGIC := '0';
    SIGNAL data_vld_in_3d             : STD_LOGIC := '0';
    
    --chaining value ulaz za blejk modul. Na pocetku svakog ip frejma vrednost je definisana shodno standardu(IV0 & IV1 & IV2 & IV3 & IV4 & IV5 & IV6 & IV7;)
    --Videti pak.vhd za tacnu vrednost. Svaki sledeci poziv koristi hash prethodnog kao vrednost. Primer: treba da odradimo hash poruke od 128 bajta. To cemo
    --uraditi uz pomoc dva poziva blejk funkcije pri cemu cemo na prvom pozivu za h_input_data staviti defaultnu vrednost a za vrednost h_input_data drugog poziva izracunati
    --hash prvog poziva.
    SIGNAL h_input_data               : STD_LOGIC_VECTOR (WIDTHx8-1  DOWNTO 0) := (OTHERS => '0'); 
    SIGNAL h_input_data_int           : STD_LOGIC_VECTOR (WIDTHx8-1  DOWNTO 0) := (OTHERS => '0'); 
    --message ulaz za blejk modul. On je realizovan kao shift registar u levo. Kada je packet_valid jednak jedinici
    --na mestu LSB bajta se upisuje packet_data i istovremeno svi ostali bajtovi siftuju za jedan bajt ulevo. Na ovaj nacin se primprema
    --64 bajtna poruka za blejk modul
    SIGNAL m_input_data               : STD_LOGIC_VECTOR (WIDTHx16-1 DOWNTO 0) := (OTHERS => '0'); 
    --salt za blejk modul. Hard wired vrednost na nulu.
    SIGNAL s_input_data               : STD_LOGIC_VECTOR (WIDTHx4-1  DOWNTO 0) := (OTHERS => '0');
    --counter promenljiva za blejk modul racunana shodno standardu   
    SIGNAL t_input_data               : STD_LOGIC_VECTOR (WIDTHx2-1  DOWNTO 0) := (OTHERS => '0'); 
    SIGNAL t_input_data_int           : STD_LOGIC_VECTOR (WIDTHx1-1  DOWNTO 0) := (OTHERS => '0'); 
    --enable za blejk modul kao znak da su podaci na ulazu validni. On se generise u dva slucaja. Prvi slucaj da je shift registar
    --m_input_data napunjen sa 64 bajtova ulaza drugi slucaj je da je padding blok zavrsio dodavanje bajtova i da su podaci spremni
    --za blejk poziv.
    SIGNAL blake_en                   : STD_LOGIC := '0';
    SIGNAL blake_en_int               : STD_LOGIC := '0';
    
    --Vrednost hash-a posle blake obrade
    SIGNAL blake_hash,blake_hash_reg  : STD_LOGIC_VECTOR (WIDTHx8-1  DOWNTO 0) := (OTHERS => '0'); 
    SIGNAL hash_ready_o_next          : STD_LOGIC := '0';
    SIGNAL blake_hash_valid           : STD_LOGIC := '0';
    SIGNAL blake_hash_valid_d         : STD_LOGIC := '0';
    
    --brojac koji broji koliko je blejk poziva odradjeno. Na pocetku svakog ip frejma se ovaj brojac resetuje. 
    --blake_cnt_max odredjuje do koliko ovaj brojac treba maksimalno da broji za trenutni ip frejm.
    SIGNAL blake_cnt, blake_cnt_max   : STD_LOGIC_VECTOR(6 DOWNTO 0) := (OTHERS => '0'); 
    SIGNAL blake_cnt_max_latched      : STD_LOGIC_VECTOR(6 DOWNTO 0) := (OTHERS => '0'); 
     
    --izlaz iz ip frame storing bloka gde je azuriran heder novim vrednostima(bajtovi za checksumu i bajtovi za duzinu poruke)
    SIGNAL start_out_int              : STD_LOGIC := '0';
    SIGNAL output_enable_int          : STD_LOGIC := '0';

    --kada se zavrsi poslednji blejk poziv njegov hash je hash izlazne poruke. Posto je hash 32-bajtni a nas izlaz bajtovski
    --koristeci sift registar u levo potrebno je isiftati njegovu vrednost. Kada je hash poslednjeg blejk poziva zavrsen cnt_hash_shifting
    --se siduje na 31 i on broji unazad do nule za to vreme se podaci isiftavaju napolje.
    SIGNAL cnt_hash_shifting          : STD_LOGIC_VECTOR(4 DOWNTO 0) := (OTHERS => '0');
    SIGNAL data_out_int               : STD_LOGIC_VECTOR(7 DOWNTO 0) := (OTHERS => '0');
    
    --padding koji se upisuje u sift registre za podatke m_input_data kada su svi bajtovi sa ulaza zavrsili upis.
    SIGNAL padding                    : STD_LOGIC_VECTOR(7 DOWNTO 0) := (OTHERS => '0');
    SIGNAL padding_valid              : STD_LOGIC := '0';
    --signal padding bloku da pocne da generise padding bajtove
    SIGNAL start_padding              : STD_LOGIC := '0';
    SIGNAL blake_en_pad               : STD_LOGIC:= '0';
    
    --ukupna duzina podatka koji se hesira u bitima potrebna za generisanje counter signala
    SIGNAL packet_length: STD_LOGIC_VECTOR(15 DOWNTO 0) := (OTHERS => '0');
    --ukupna duzina paddinga u bajtovima neophodna za odredjivanje kasnjenja kod adresibilnih sift registara
    SIGNAL pading_length_int: STD_LOGIC_VECTOR(6 DOWNTO 0) := (OTHERS => '0');
    
    
    --pomocni signal koji definise da li je posle paddinga potrebno odraditi jedan ili dva blejk poziva. Ukoliko je vrednost padding-a veca od 64 bajtova potrebna
    --su dva dodatna blejk poziva.
    SIGNAL two_blake: STD_LOGIC := '0';
    
    
    --Ovaj signal se podize na jedinicu u trenutku kada je zavrsen prijem trenutnog frejma kao signal
    --ulazu da sledeci frejm jos ne moze da se prihvati.
    SIGNAL busy                       : STD_LOGIC := '0';
    --ovaj counter pocinje da broji kada se busy signal aktivira.
    SIGNAL cnt_busy                   : STD_LOGIC_VECTOR(6 DOWNTO 0) := (OTHERS => '0');
    --vrednost granicne vrednosti busy countera. Ova vrednost definise dokle ce counter brojati
    --i kada counter dobroji do ove vrednosti busy signal se deaktivira. Izmedju dva frejma mora postojati pauza 
    --od bar 107 taktova tako da je vrednost ovog limita upravo ta.
    SIGNAL cnt_busy_upper_limit       : STD_LOGIC_VECTOR(6 DOWNTO 0) := (OTHERS => '0');
    
BEGIN 

    --generisanje busy signala u trenutku kada poslednji bajt ip poruke pristize
    busy <= '1' WHEN ((cnt=cnt_upper_limit_busy AND data_vld_in = '1') OR cnt = cnt_upper_limit) ELSE '0';
    
    --dok je busy flag aktivan modul nije spreman da prihvati novu poruku
    ready_to_accept <= not busy; 
    
    --brojac taktova busy flag-a
    PROCESS(clk)
    BEGIN
        IF (clk'EVENT AND clk = '1') THEN
            IF(start_in = '1' AND data_vld_in = '1') THEN
                cnt_busy <= (OTHERS => '0');
            ELSIF(busy = '1') THEN
                cnt_busy <= cnt_busy + conv_std_logic_vector(1,7);
            END IF;
        END IF;
    END PROCESS;

    
    PROCESS(clk) 
    BEGIN
        IF (clk'EVENT AND clk = '1') THEN
        
            --posle 107 taktova pauze modul je spreman da prihvati novu ip poruku nezavisno od njene duzine.
            cnt_busy_upper_limit <= conv_std_logic_vector(107,7);

            --Brojac koji broji pristigle podatke. Na pocetku svakog novog frejma vrednost brojaca se resetuje.
            IF(cnt_busy = cnt_busy_upper_limit) THEN
                cnt <= (OTHERS => '0');            
            ELSIF(start_in = '1' AND data_vld_in = '1') THEN
                cnt <= (OTHERS => '0');
            ELSIF(data_vld_in = '1') THEN
                cnt <= cnt + conv_std_logic_vector(1,11);
            END IF;
            
        END IF;
    END PROCESS;

    PROCESS(clk)
    BEGIN
        IF (clk'EVENT AND clk = '1') THEN
            data_vld_in_d <= data_vld_in;
            data_vld_in_2d <= data_vld_in_d;
            data_vld_in_3d <= data_vld_in_2d;
  
            data_in_d <= data_in;         
            data_in_2d <= data_in_d;
            data_in_3d <= data_in_2d;
                
            cnt_d <= cnt;   
            cnt_2d <= cnt_d; 
            cnt_3d <= cnt_2d; 
        END IF;
    END PROCESS;

    PROCESS(clk)
    BEGIN
        IF (clk'EVENT AND clk = '1') THEN
            --cuvanje duzine ip poruke
            IF(data_vld_in = '1' AND cnt = conv_std_logic_vector(1,11)) THEN --MSB bajt od ukupne duzine
                total_length(15 DOWNTO 8) <= data_in;
            ELSIF(data_vld_in = '1' AND cnt = conv_std_logic_vector(2,11)) THEN --LSB bajt od ukupne duzine
                total_length(7 DOWNTO 0) <= data_in;
            END IF;
            
            --gornje vrednosti countera koji broji pristigle bajtove
            cnt_upper_limit <= total_length(10 DOWNTO 0) - conv_std_logic_vector(1,11);
            cnt_upper_limit_busy <= total_length(10 DOWNTO 0) - conv_std_logic_vector(2,11);
            
            --nova duzina poruke uvecana za duzinu hash-a
            total_new_length <= total_length + conv_std_logic_vector(32,16);
        END IF;
    END PROCESS;

    --Masina za racunanje check sume nadolazeceg IP hedera
    PROCESS(clk)
    BEGIN
        IF (clk'EVENT AND clk = '1') THEN
            IF(start_in = '1') THEN
                check_sum_aux_2 <= (OTHERS => '0');
            ELSE
                check_sum_aux_2 <= check_sum_aux_1(15 DOWNTO 0) + check_sum_aux_1(16);
            END IF;
        
            IF(start_in = '1' AND data_vld_in = '1') THEN
                state_of_receiving_byte <= state_LSB;
                check_sum_aux_1<= (OTHERS => '0');
                latched_msb_data <= data_in;
            ELSIF(data_vld_in = '1') THEN
                CASE state_of_receiving_byte IS
                    WHEN state_MSB =>
                        state_of_receiving_byte <= state_LSB;
                        latched_msb_data <= data_in;
                    WHEN state_LSB =>
                        state_of_receiving_byte <= state_MSB;
                        check_sum_aux_1 <= ('0' & check_sum_aux_2) + ('0'  &  latched_msb_data  &  data_in);
                    WHEN OTHERS =>
                END CASE;
            END IF;

        END IF;
    END PROCESS;

    check_sum <= not check_sum_aux_2;

    PROCESS(clk)
    BEGIN
        IF (clk'EVENT AND clk = '1') THEN
            IF(start_in = '1' and data_vld_in = '1') THEN --fleg koji definise da li je ip paket validan se resetuje na pocetku
                invalid_frame <= '0';
            ELSIF(data_vld_in = '1' AND cnt = conv_std_logic_vector(0,11) AND data_in_d(7 DOWNTO 4) /= VERSION_IPV4) THEN --Nije IPv4 protokol. Takav frame nije validan.
                invalid_frame <= '1';
            ELSIF(data_vld_in_2d = '1' AND check_sum /= conv_std_logic_vector(0,16) AND cnt_d = conv_std_logic_vector(19,11)) THEN --Losa checksuma. Odbaci frame.
                invalid_frame <= '1';
            END IF;
        END IF;
    END PROCESS;
    
    
    
    PROCESS(clk)
    BEGIN
        IF (clk'EVENT AND clk = '1') THEN
        
            --Ukoliko je frejm validan payload frejma se prosledjuje blake modulu
            IF(data_vld_in_3d = '1' AND invalid_frame = '0' AND cnt_2d >= conv_std_logic_vector(20,11)) THEN 
                packet_valid <= '1';
                packet_data <= data_in_3d;
            ELSE
                packet_valid <= '0';            
            END IF;
            
            --na kraju ip poruke potrebno je dodati padding bajtove. start_padding signal javlja padding_generation bloku da krene da generise padding bajtove.
            IF(cnt_2d = cnt_upper_limit AND  data_vld_in_3d = '1' AND invalid_frame = '0') THEN
                start_padding <= '1';
            ELSE
                start_padding <= '0';
            END IF;            
            
        END IF;
    END PROCESS;
    
    --Instanca padding_generation bloka
    inst_padding_generation: padding_generation
    GENERIC MAP(
        HEDER_LENGTH => HEDER_LENGTH
    )
    PORT MAP (
        clk => clk,
        --ulazni interfejs
        total_length => total_length,
        start_padding => start_padding,
        --izlazni interfejs
        packet_length => packet_length,
        pading_length => pading_length_int,
        two_blake => two_blake,
        blake_en => blake_en_pad,
        blake_cnt_max => blake_cnt_max,
        padding => padding,
        padding_valid => padding_valid
    );
    
    
    --salt je hard-wired na nulu i u vhdl modulu i u matlab referentnoj simulaciji
    s_input_data <= (OTHERS => '0');
    
    
    --message za blake modul realizovan kao sift registar u levo. Podaci iz ip poruke ili padding
    --se upisuju na LSB mesto i istovremeno usiftavaju kako bi se podaci pripremili za blake modul
    PROCESS(clk)
    BEGIN
        IF (clk'EVENT AND clk = '1') THEN
            IF(packet_valid = '1') THEN --ip podaci
                m_input_data(7 DOWNTO 0) <= packet_data;
                m_input_data(WIDTHx16-1 DOWNTO 8) <= m_input_data(WIDTHx16-9 DOWNTO 0);  
            ELSIF(padding_valid = '1') THEN --padding podaci
                m_input_data(7 DOWNTO 0) <= padding;
                m_input_data(WIDTHx16-1 DOWNTO 8) <= m_input_data(WIDTHx16-9 DOWNTO 0);              
            END IF;  
        END IF;
    END PROCESS;

  
    --Enable za blake modul kada je prikupljeno dovoljno bajtova sa ulaza
    PROCESS(clk)
    BEGIN
       IF (clk'EVENT AND clk = '1') THEN
           IF((packet_valid = '1' AND cnt_3d(5 DOWNTO 0) = "010011")) THEN 
               blake_en_int <= '1';
           ELSE
               blake_en_int <= '0';            
           END IF;  
       END IF;
    END PROCESS;   
    
    PROCESS(clk)
    BEGIN
        IF (clk'EVENT AND clk = '1') THEN
            --chain vrednost za blake modul
            IF(packet_valid = '1' AND cnt_3d = "00001010011") THEN
                h_input_data_int <= IV0 & IV1 & IV2 & IV3 & IV4 & IV5 & IV6 & IV7;            
            ELSIF (start_out_int = '1' AND output_enable_int = '1') THEN
                h_input_data_int <= IV0 & IV1 & IV2 & IV3 & IV4 & IV5 & IV6 & IV7;
            ELSIF(hash_ready_o_next = '1') THEN --sacuvajmo novu chain vrednost kada blake modul zavrsi procesiranje
                h_input_data_int <= blake_hash;            
            END IF; 

            --sacuvajmo chain vrednost i usiftavamo u levo kako bi ti podaci isli na izlaz (krajnji hash)
            IF(hash_ready_o_next = '1') THEN
                blake_hash_reg <= blake_hash;  
            ELSIF(cnt_hash_shifting /= conv_std_logic_vector(0,5)) THEN
                blake_hash_reg(WIDTHx8-1  DOWNTO 8) <= blake_hash_reg(WIDTHx8-1-8  DOWNTO 0);
            END IF; 
        
        END IF;
    END PROCESS;   

    --ovo je slucaj kada je poruka manja od 64 bajtova i mora da se odradi padding...potrebno je ranije resetovati inicijalnu chain vrednost
    h_input_data <= IV0 & IV1 & IV2 & IV3 & IV4 & IV5 & IV6 & IV7 WHEN (blake_en_pad = '1' and blake_cnt = conv_std_logic_vector(0,7)) ELSE h_input_data_int;

    --counter promenljiva za blake funkciju
    t_input_data (31  DOWNTO 0) <= (OTHERS => '0');
    
    --counter vrednost za blake poziv ciji message u sebi sadrzi padding bajtove
    t_input_data (63  DOWNTO 32) <= "0000000000000" & packet_length & "000" WHEN (blake_en_pad = '1' and padding_valid = '1' and two_blake = '1') ELSE --ovo je slucaj kada je broj padding bajtova veci od 65 pa su dodatna dva blake poziva neophodna..ovo je vrednost countera prilikom prvog poziva
                                    "0000000000000" & packet_length & "000" WHEN (blake_en_pad = '1' and padding_valid = '0' and two_blake = '0') ELSE --ovo je slucaj kada je jedan dodatni blake poziv potreban i vrednost counter promenljive za taj slucaj
                                     X"00000000" WHEN (blake_en_pad = '1' and padding_valid = '0' and two_blake = '1') ELSE -- ovo je slucaj kada je broj padding bajtova veci od 65 pa su dodatna dva blake poziva neophodna...za drugi blake poziv counter promenljiva ima vrednost nula po specifikaciji
                                     t_input_data_int;
    
    PROCESS(clk)
    BEGIN
        IF (clk'EVENT AND clk = '1') THEN
        
            --counter promenljiva za blake pozive koji ne sadrze padding bajtove u message promenljivoj
            IF(packet_valid = '1' AND cnt_3d = "00001010011") THEN
                t_input_data_int <= conv_std_logic_vector(512,32);
            ELSIF(packet_valid = '1' AND cnt_3d(5 DOWNTO 0) = "010011") THEN
                t_input_data_int <= t_input_data (63  DOWNTO 32) + conv_std_logic_vector(512,32);
            END IF;
       
            --counter koji broji blake pozive u jednom ip frejmu 
            IF(blake_hash_valid = '0'  AND blake_hash_valid_d = '1') THEN
                blake_cnt <= (OTHERS => '0');
            ELSIF(hash_ready_o_next = '1') THEN
                blake_cnt <= blake_cnt + conv_std_logic_vector(1,7);
            END IF;  

        END IF;
    END PROCESS;

    PROCESS(clk)
    BEGIN
        IF (clk'EVENT AND clk = '1') THEN
        
            --sacuvajmo koliko blake poziva je neophodno za trenutni ip frejm
            IF (start_out_int = '1' AND output_enable_int = '1') THEN 
                blake_cnt_max_latched <= blake_cnt_max;
            END IF;
   
            -- kada se poslednji blake poziv zavrsi u narednih 32 takta se chain vrednost salje na izlaz
            IF(blake_cnt = blake_cnt_max_latched and hash_ready_o_next = '1') THEN
                cnt_hash_shifting <= (OTHERS => '1');
            ELSIF(cnt_hash_shifting /= conv_std_logic_vector(0,5)) THEN
                cnt_hash_shifting <= cnt_hash_shifting - conv_std_logic_vector(1,5);
            END IF;

            --ukoliko je poslednji blake poziv u pitanju generise se pulse
            IF(cnt_hash_shifting /= conv_std_logic_vector(0,5) or (blake_cnt = blake_cnt_max_latched and hash_ready_o_next = '1')) THEN
                blake_hash_valid <= '1';
            ELSE
                blake_hash_valid <= '0';
            END IF;
            
            blake_hash_valid_d <= blake_hash_valid;
            
        END IF;
    END PROCESS;
    
    
    --instanca blake funkcije
    blake_pipe_com: blake_pipe
    PORT MAP (
        clk => clk ,
      
        h => h_input_data ,
        m => m_input_data ,
        s => s_input_data ,
        t => t_input_data ,
        inEN =>  blake_en,
        hash => blake_hash,
        outEN => hash_ready_o_next  
    );
     
    --signal dozvole za blake funkciju...javlja se ukoliko je dovoljno podataka sa ulaza prikupljeno(64 bajtova) ili ukoliko je signal iz padding generation bloka aktivan
    --signal iz padding generation bloka nakon dodatog paddinga signalizira da moze da se dozvoli blake poziv
    blake_en <= blake_en_int or blake_en_pad;
    
    inst_ip_frame_storing: ip_frame_storing
    PORT MAP(
        clk => clk,
        start_in => start_in,
        data_vld_in => data_vld_in_3d,
        data_in => data_in_3d,
        total_new_length => total_new_length,
        cnt => cnt_2d,
        pading_length => pading_length_int,
        start_out=> start_out_int,
        output_enable => output_enable_int,
        data_out => data_out_int
    );
    
    --kasnjenje invalid_frame kroz adresabilan sift registar
    invalid_frame_int_aux(0) <= invalid_frame;
    
    inst_invalid_frame_delay: addressable_shift_register
    GENERIC MAP(
        DATA_SIZE => 1,
        MAX_DELAY => 94,
        ADDR_LENGTH => 7
    )
    PORT MAP(
        clk => clk,
        data_in => invalid_frame_int_aux,
        address => address_shift_reg,
        data_out => invalid_frame_aux
    );
    
    PROCESS (clk)
    BEGIN
        IF (clk'EVENT AND clk = '1') THEN
            address_shift_reg_aux <= pading_length_int + conv_std_logic_vector(41,7);--30
            
            IF(address_shift_reg_aux > conv_std_logic_vector(93,7)) THEN--82
                address_shift_reg <= conv_std_logic_vector(93,7);--82
            ELSE
                address_shift_reg <= address_shift_reg_aux;
            END IF;
            
        END IF;
    END PROCESS;  

    --generisanje start izlaznog signala ukoliko je frame validan
    start_out <= start_out_int and (not invalid_frame_aux(0));
    
    --krajnji output enable signal
    output_enable <= (output_enable_int and (not invalid_frame_aux(0))) or blake_hash_valid;

    --krajnji izlaz je izlaz iz ip_frame_storing bloka ili sift registra u kome se cuva krajnji hash
    data_out <= blake_hash_reg(WIDTHx8-1  DOWNTO WIDTHx8-8) WHEN blake_hash_valid = '1' ELSE data_out_int;
    
END RTL;