LIBRARY IEEE;
USE IEEE.STD_LOGIC_1164.ALL;
USE IEEE.NUMERIC_STD.ALL;
USE IEEE.STD_LOGIC_UNSIGNED.ALL;
USE IEEE.STD_LOGIC_ARITH.ALL;

ENTITY padding_generation IS
    GENERIC(
        HEDER_LENGTH : INTEGER := 20             
    );
    PORT (
        clk: IN STD_LOGIC;
        --ulazni interfejs
        total_length: IN STD_LOGIC_VECTOR(15 DOWNTO 0);
        start_padding: IN STD_LOGIC;
        --izlazni interfejs
        packet_length: OUT STD_LOGIC_VECTOR(15 DOWNTO 0);
        pading_length: OUT STD_LOGIC_VECTOR(6 DOWNTO 0);
        two_blake: OUT STD_LOGIC;
        blake_en: OUT STD_LOGIC;
        blake_cnt_max: OUT STD_LOGIC_VECTOR(6 DOWNTO 0);
        padding: OUT STD_LOGIC_VECTOR(7 DOWNTO 0);
        padding_valid: OUT STD_LOGIC
    );
END padding_generation;

ARCHITECTURE RTL OF padding_generation IS

    CONSTANT MINIMUM_PADDING: INTEGER  := 9;--u bajtovima
    CONSTANT MAXIMUM_PADDING: INTEGER  := 72;--u bajtovima

    SIGNAL packet_length_int: STD_LOGIC_VECTOR(15 DOWNTO 0) := (OTHERS => '0');
    SIGNAL remainder: STD_LOGIC_VECTOR(5 DOWNTO 0) := (OTHERS => '0');

    SIGNAL l_64: STD_LOGIC_VECTOR(63 DOWNTO 0) := (OTHERS => '0'); 
    
    SIGNAL pading_length_int: STD_LOGIC_VECTOR(6 DOWNTO 0) := (OTHERS => '0');
    SIGNAL blake_num_base: STD_LOGIC_VECTOR(6 DOWNTO 0) := (OTHERS => '0');
    
    SIGNAL cnt_padding: STD_LOGIC_VECTOR(6 DOWNTO 0) := (OTHERS => '0');
    SIGNAL cnt_padding_upper_limit: STD_LOGIC_VECTOR(6 DOWNTO 0) := (OTHERS => '0');
    SIGNAL padding_valid_int: STD_LOGIC := '0';
    SIGNAL cnt_length_8: STD_LOGIC_VECTOR(6 DOWNTO 0) := (OTHERS => '0');
    SIGNAL cnt_length_7: STD_LOGIC_VECTOR(6 DOWNTO 0) := (OTHERS => '0');
    SIGNAL cnt_length_6: STD_LOGIC_VECTOR(6 DOWNTO 0) := (OTHERS => '0');
    SIGNAL cnt_length_5: STD_LOGIC_VECTOR(6 DOWNTO 0) := (OTHERS => '0');
    SIGNAL cnt_length_4: STD_LOGIC_VECTOR(6 DOWNTO 0) := (OTHERS => '0');
    SIGNAL cnt_length_3: STD_LOGIC_VECTOR(6 DOWNTO 0) := (OTHERS => '0');
    SIGNAL cnt_length_2: STD_LOGIC_VECTOR(6 DOWNTO 0) := (OTHERS => '0');
    SIGNAL cnt_length_1: STD_LOGIC_VECTOR(6 DOWNTO 0) := (OTHERS => '0');
    SIGNAL cnt_length_0: STD_LOGIC_VECTOR(6 DOWNTO 0) := (OTHERS => '0');    

BEGIN
    --ostatak pri deljenju ukupne duzine paketa bez hedera sa 64. Ovaj broj je znacajan kako bi odredili koliko bajtova za padding je neophodno.
    remainder <= packet_length_int(5 DOWNTO 0);
    --Ukupan broj blake poziva bez padding-a.
    blake_num_base <= packet_length_int(12 DOWNTO 6);
    --ukupan broj bita koji prolazi kroz lanac blake funkcija. Ova promenljiva je bitna kod padding-a.
    l_64(18 downto 0) <= packet_length_int & "000";
    l_64(63 DOWNTO 19) <= (others => '0');
    
    pading_length <= pading_length_int;
    packet_length <= packet_length_int;
   
    -- ukupna duzina paketa bez hedera. Nova vrednost se ne azurira ako je padding.
    PROCESS(clk)
    BEGIN
        IF (clk'EVENT AND clk = '1') THEN
            IF(padding_valid_int /= '1') THEN
                packet_length_int <= total_length - conv_std_logic_vector(HEDER_LENGTH,16);
            END IF;
        END IF;
    END PROCESS;
    
    
    --Odredjivanje duzine padding-a ukoliko je neophodan kao i ukupan broj blake poziva sa paddingom.
    PROCESS(clk)
    VARIABLE  aux: STD_LOGIC_VECTOR(7 DOWNTO 0);
    BEGIN
        IF (clk'EVENT AND clk = '1') THEN
            IF(remainder = conv_std_logic_vector(0,6)) THEN --Nije neophodan padding
                pading_length_int <= (others => '0');
                blake_cnt_max <= blake_num_base - conv_std_logic_vector(1,7);
                two_blake <= '0';
            ELSIF(remainder > conv_std_logic_vector(64-MINIMUM_PADDING,6)) THEN --potrebna su dodatna dva blake poziva posle padding-a.
                aux := conv_std_logic_vector(128,8) - ("00" & remainder);
                pading_length_int <= aux(6 DOWNTO 0);
                blake_cnt_max <= blake_num_base + conv_std_logic_vector(1,7);
                two_blake <= '1';
            ELSE  --jedan dodatni blake poziv posle padding-a
                pading_length_int <= conv_std_logic_vector(64,7) - ('0' & remainder);
                blake_cnt_max <= blake_num_base;
                two_blake <= '0';
            END IF;
            cnt_padding_upper_limit <= pading_length_int - conv_std_logic_vector(1,7);--vrednost gornje granice countera koji sluzi za kontrolu generisanja padding-a
        END IF;
    END PROCESS;
    
    --Kontrola logika za generisanje padding-a
    PROCESS(clk)
    BEGIN
        IF (clk'EVENT AND clk = '1') THEN
            --ukoliko stigne zahtev za generisanje padding-a resetuje se counter i postavlja valid za padding u aktivno stanje
            IF(start_padding = '1' and pading_length_int/= conv_std_logic_vector(0,7)) THEN
                cnt_padding <= (others => '0');
                padding_valid_int <= '1';
            ELSIF(cnt_padding = cnt_padding_upper_limit and padding_valid_int = '1') THEN --kad brojac izbroji do kraja valid za padding se postavlja u neaktivno stanje 
                padding_valid_int <= '0';
            ELSE --u ostalim slucajevima counter broji
                cnt_padding <= cnt_padding + conv_std_logic_vector(1,7);
            END IF;
            
            
            --Vrednost counter promenljive za specificne bajtove padding-a. 
            cnt_length_8 <= cnt_padding_upper_limit - conv_std_logic_vector(9,7);
            cnt_length_7 <= cnt_padding_upper_limit - conv_std_logic_vector(8,7);
            cnt_length_6 <= cnt_padding_upper_limit - conv_std_logic_vector(7,7);
            cnt_length_5 <= cnt_padding_upper_limit - conv_std_logic_vector(6,7);
            cnt_length_4 <= cnt_padding_upper_limit - conv_std_logic_vector(5,7);
            cnt_length_3 <= cnt_padding_upper_limit - conv_std_logic_vector(4,7);
            cnt_length_2 <= cnt_padding_upper_limit - conv_std_logic_vector(3,7);
            cnt_length_1 <= cnt_padding_upper_limit - conv_std_logic_vector(2,7);
            cnt_length_0 <= cnt_padding_upper_limit - conv_std_logic_vector(1,7);
            
            --Kad se zavrsi generisanje padding-a signal za dozvolu generisanja blake funkcije se postavlja u aktivno stanje 
            IF(cnt_padding(5 DOWNTO 0) = cnt_padding_upper_limit(5 DOWNTO 0) and padding_valid_int = '1') THEN
                blake_en <= '1';
            ELSE
                blake_en <= '0';
            END IF;
            
            
            
            IF(start_padding = '1') THEN --prvi bajt padding-a
                IF (cnt_padding_upper_limit = conv_std_logic_vector(MINIMUM_PADDING-1,7)) THEN --ovo je slucaj da je potrebno dodati 9 bajtova padding-a. U tom slucaju prvi bajt je X"81"
                    padding <= X"81";
                ELSE --u svim ostalim slucajevima prvi bajt je X"80"
                    padding <= X"80";
                END IF;
            ELSIF(cnt_padding = cnt_length_8) THEN --bajt koji se salje neposredno pre ukupnog broja bita blake funkcije je fiksan i iznosi X"01"
                padding <= X"01";
            ELSIF(cnt_padding = cnt_length_7) THEN --MSB od broja bita koji prolaze kroz blake chain
                padding <= l_64(8*8-1 DOWNTO 8*7); 
            ELSIF(cnt_padding = cnt_length_6) THEN  
                padding <= l_64(8*7-1 DOWNTO 8*6);            
            ELSIF(cnt_padding = cnt_length_5) THEN 
                padding <= l_64(8*6-1 DOWNTO 8*5);            
            ELSIF(cnt_padding = cnt_length_4) THEN  
                padding <= l_64(8*5-1 DOWNTO 8*4);            
            ELSIF(cnt_padding = cnt_length_3) THEN  
                padding <= l_64(8*4-1 DOWNTO 8*3);             
            ELSIF(cnt_padding = cnt_length_2) THEN  
                padding <= l_64(8*3-1 DOWNTO 8*2);            
            ELSIF(cnt_padding = cnt_length_1) THEN  
                padding <= l_64(8*2-1 DOWNTO 8*1);           
            ELSIF(cnt_padding = cnt_length_0) THEN     
                padding <= l_64(8*1-1 DOWNTO 8*0); --LSB bajt od broja bita koji prolaze kroz blake chain
            ELSE
                padding <= (OTHERS => '0'); -- svi bajtovi osim prvog i poslednjih devet su jednaki nuli
            END IF;    
        END IF;
    END PROCESS;
    
    padding_valid <= padding_valid_int;

END RTL;